# content of conftest.py
import pytest
import logging as log
import pandas as pd

from src.chainlink_feeds import ChainlinkFeeds
log.basicConfig(level=log.INFO)


@pytest.fixture
def chainlink_feed():
    cf = ChainlinkFeeds()
    return cf


def test_convert_pair_format():
    # Arrange
    unconverted_pair = 'ETH_USD'

    # Act
    converted_pair = ChainlinkFeeds.convert_pair_format(unconverted_pair)

    # Assert
    assert converted_pair == 'ETH/USD'


def test_get_prices(chainlink_feed):
    # Arrange
    cf = chainlink_feed

    # Act
    result = cf.get_prices(pair='eth_usd')

    # Assert
    result[0]['assetPair'] == "ETH/USD"


def test_get_latest_round_data(chainlink_feed):
    # Arrange
    cf = chainlink_feed
    pair = 'eth_usd'

    # Act
    result = cf.get_latest_round_data(pair=pair)

    # Assert
    result[0]['assetPair'] == "ETH/USD"


def test_graphql_query(chainlink_feed):
    # Arrange
    cf = chainlink_feed
    query = """{
                    prices(where: {assetPair:"LINK/ETH"}, orderBy: timestamp, orderDirection:desc, first: 1000) {
                    id
                    price
                    timestamp
                    blockNumber
                    blockHash
                    transactionHash
                    assetPair
                    }
                }"""
    # Act
    result = cf.graphql_query(query)

    # Assert
    assert result['prices'][0]['assetPair'] == "LINK/ETH"


def test_pandas_output():
    # Arrange
    cf = ChainlinkFeeds(output_format='pandas')

    # Act
    result = cf.get_prices(pair='eth_usd')

    # Asset
    assert isinstance(result, pd.DataFrame)
    assert result['assetPair'][0] == 'ETH/USD'


def test_get_price_feeds(chainlink_feed):
    # Arrange
    cf = chainlink_feed

    # Act
    result = cf.get_price_feeds()

    # Assert
    assert isinstance(result[0]['assetPair'], str)


def test_get_hourly_candle(chainlink_feed):
    # Arrange
    cf = chainlink_feed

    # Act
    result = cf.get_hourly_candle(pair='ETH_USD')

    # Assert
    assert isinstance(result[0]['assetPair'], str)


def test_get_daily_candle(chainlink_feed):
    # Arrange
    cf = chainlink_feed

    # Act
    result = cf.get_daily_candle(pair='ETH_USD')

    # Assert
    assert isinstance(result[0]['assetPair'], str)


def test_get_weekly_candle(chainlink_feed):
    # Arrange
    cf = chainlink_feed

    # Act
    result = cf.get_weekly_candle(pair='ETH_USD')

    # Assert
    assert isinstance(result[0]['assetPair'], str)



import os
from web3 import Web3
import configparser
import logging as log
from functools import wraps
from datetime import datetime
import json
import requests
import pandas as pd
log.basicConfig(level=log.INFO)


def _rpc_url_formatter(func):
    @wraps(func)
    def _conversion_formating(self, *args, **kwargs):
        result_dict = func(self, *args, **kwargs)
        result_dict['price'] = float(self.web3.fromWei(
            result_dict['price'], self.conversion))
        if self.time is not None:
            result_dict['time_stamp'] = datetime.fromtimestamp(
                result_dict['time_stamp']).strftime(self.time)
            result_dict['started_at'] = datetime.fromtimestamp(
                result_dict['started_at']).strftime(self.time)
        return result_dict
    return _conversion_formating


# def _handle_api_call(func):
#     @wrape
class ChainlinkFeeds(object):
    """ Base class where the decorators and base function for the other
    classes of this python wrapper will inherit from.

    TODO: Add The Graph for historical
    """

    def __init__(self, rpc_url=None, output_format='json', proxy=None, rpc_env_var=None, conversion='ether',
                 time='%Y-%m-%d %H:%M:%S'):
        """ Initialize the class
        Keyword Arguments:
            key:  rpc_url: The Ethereum Client node. This can be from node hosting services like infura, fiews, or quiknode.
            output_format:  Either 'json' or 'pandas'
            proxy:  Dictionary mapping protocol or protocol and hostname to the URL of the proxy.
        """
        self.rpc_url = rpc_url
        if rpc_url is not None or rpc_env_var is not None:
            if rpc_env_var is None:
                self.rpc_url = os.getenv('RPC_URL')
            else:
                self.rpc_url = os.getenv(rpc_env_var)
            if not self.rpc_url or not isinstance(self.rpc_url, str):
                raise ValueError(
                    'No value RPC_URL provided. Please set environment variable RPC_URL, ',
                    'provide an rpc_url as a keyword argument, or set your own environment',
                    ' variable name with rpc_env_var.')
            else:
                log.info(
                    'You have chosen the RPC_URL. You will only be able to get specific historical data. ')
        self.abis_config = configparser.ConfigParser()
        self.addresses_config = configparser.ConfigParser()
        self.load_config(os.path.join(os.path.dirname(
            __file__), 'config', 'addresses.cfg'), abi_or_address='address')
        self.load_config(os.path.join(os.path.dirname(
            __file__), 'config', 'abis.cfg'), abi_or_address='abi')
        if self.rpc_url:
            self.web3 = Web3(Web3.HTTPProvider(self.rpc_url))
        self.proxy = proxy or {}
        self.output_format = output_format
        self.conversion = conversion
        self.time = time
        self.base_url = 'https://api.thegraph.com/subgraphs/name/melonproject/chainlink'

    def get_latest_round_data(self, network='kovan', pair='eth_usd', address=None, abi=None):
        """Pairs that end with ETH have 18 0s. All other pairs only have 10.
        We multiply them by 'bump' so they are all in wei before conversion.
        """
        if self.rpc_url is not None:
            return self.get_latest_round_data_rpc(network=network, pair=pair, address=address, abi=abi)
        return self.get_prices(pair=pair, number_of_results=1)

    def get_historical_price(self, round_id=None, network='kovan', pair='eth_usd', address=None, abi=None):
        """Pairs that end with ETH have 18 0s. All other pairs only have 10.
        We multiply them by 'bump' so they are all in wei before conversion.
        """
        if self.rpc_url is not None:
            return self.get_historical_price_rpc(round_id=round_id, network=network, pair=pair, address=address, abi=abi)
        return self.get_prices(pair=pair, round_id=round_id)

    def get_price_feeds(self, first=1000):
        query = """{{
            priceFeeds(first: {first}) {{
                id
                assetPair
            }}
        }}""".format(first=first)
        return self.graphql_query(query, column='priceFeeds')

    def get_prices(self, pair='eth_usd', number_of_results=1000, round_id=None):
        # TODO round_id
        pair = ChainlinkFeeds.convert_pair_format(pair)
        query = """{{
                    prices(where: {{assetPair:"{pair}"}}, orderBy: timestamp,
                        orderDirection:desc, first: {number_of_results}) {{
                    id
                    price
                    timestamp
                    blockNumber
                    blockHash
                    transactionHash
                    assetPair
                    }}
                }}""".format(pair=pair, number_of_results=number_of_results)
        return self.graphql_query(query, column='prices', set_index='timestamp')

    def get_hourly_candle(self, pair='eth_usd', number_of_results=1000):
        pair = ChainlinkFeeds.convert_pair_format(pair)
        query = ChainlinkFeeds.get_candle_query(
            pair, number_of_results, 'hourlyCandles')
        return self.graphql_query(query, column='hourlyCandles', set_index='openTimestamp')

    def get_daily_candle(self, pair='eth_usd', number_of_results=1000):
        pair = ChainlinkFeeds.convert_pair_format(pair)
        query = ChainlinkFeeds.get_candle_query(
            pair, number_of_results, 'dailyCandles')
        return self.graphql_query(query, column='dailyCandles', set_index='openTimestamp')

    def get_weekly_candle(self, pair='eth_usd', number_of_results='1000'):
        pair = ChainlinkFeeds.convert_pair_format(pair)
        query = ChainlinkFeeds.get_candle_query(
            pair, number_of_results, 'weeklyCandles')
        return self.graphql_query(query, column='weeklyCandles', set_index='openTimestamp')

    @staticmethod
    def get_candle_query(pair, number_of_results, endpoint):
        return """{{{endpoint}(where: {{assetPair: "{pair}" }}, orderBy: openTimestamp, orderDirection: desc, first: {number_of_results}) {{
                    assetPair
                    openTimestamp
                    closePrice
                    highPrice
                    lowPrice
                    openPrice
                    closePrice
                    averagePrice
                    medianPrice
                }}
                }}""".format(pair=pair, number_of_results=number_of_results, endpoint=endpoint)

    def graphql_query(self, query, column=None, set_index=None):
        response = requests.post(self.base_url, json={'query': query})
        return self._handle_output(json.loads(response.text), column=column, set_index=set_index)

    def _handle_output(self, response, column=None, set_index=None):
        if self.output_format == 'pandas':
            if column:
                df = pd.DataFrame(response['data'][column])
            else:
                df = pd.DataFrame(response['data'])
            if set_index:
                df = df.set_index(set_index)
            return df
        else:
            if column:
                return response['data'][column]
            return response['data']

    @staticmethod
    def convert_pair_format(pair):
        pair = pair.upper()
        if "_" in pair:
            return pair.replace("_", "/")
        return pair

    """
    The functions below are ment specifically for those who want to use the RPC_URL instead of the query from The Graph.
    """

    def load_config(self, config_path, abi_or_address='abi'):
        if abi_or_address.lower() == 'abi':
            self.abis_config.read(config_path)
        else:
            self.addresses_config.read(config_path)

    def get_manual_addresses(self):
        return json.loads(json.dumps(self.addresses_config._sections))

    def get_manual_abis(self):
        return json.loads(json.dumps(self.abis_config._sections))

    @_rpc_url_formatter
    def get_historical_price_rpc(self, round_id, network='kovan', pair='eth_usd', address=None, abi=None):
        """Pairs that end with ETH have 18 0s. All other pairs only have 10.
        We multiply them by 'bump' so they are all in wei before conversion.
        """
        bump = 1
        if not pair.endswith('eth'):
            bump = 10000000000
        if address is None:
            address = self.addresses_config[network.lower()][pair]
        if abi is None:
            abi = self.abis_config['default']['aggregatorv3interfaceabi']
        price_feed_contract = self.web3.eth.contract(address=address, abi=abi)
        latest_data = price_feed_contract.functions.getRoundData(
            round_id).call()
        result_dict = {'round_id': latest_data[0], 'price': latest_data[1] * bump,
                       'started_at': latest_data[2], 'time_stamp': latest_data[3], 'answered_in_round': latest_data[4]}
        return result_dict

    @_rpc_url_formatter
    def get_latest_round_data_rpc(self, network='kovan', pair='eth_usd', address=None, abi=None):
        """Pairs that end with ETH have 18 0s. All other pairs only have 10.
        We multiply them by 'bump' so they are all in wei before conversion.
        """
        bump = 1
        if not pair.endswith('ETH'):
            bump = 10000000000
        if address is None:
            address = self.addresses_config[network.lower()][pair]
        if abi is None:
            abi = self.abis_config['default']['aggregatorv3interfaceabi']
        price_feed_contract = self.web3.eth.contract(address=address, abi=abi)
        latest_data = price_feed_contract.functions.latestRoundData().call()
        result_dict = {'round_id': latest_data[0], 'price': latest_data[1] * bump,
                       'started_at': latest_data[2], 'time_stamp': latest_data[3], 'answered_in_round': latest_data[4]}
        return result_dict



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
short_description = 'Python module to get price and other data from the decentralized chainlink community resources'

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    # How you named your package folder (MyLib)  # Chose the same as "name"
    name='chainlink_feeds',
    version='0.2.7',      # Start with a small number and increase it with every change you make
    # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    license='MIT',
    # Give a short description about your library
    description=short_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Patrick Collins',                   # Type in your name
    author_email='patrick@alphachain.io',      # Type in your E-Mail
    # Provide either the link to your github or to your website
    url='https://github.com/alphachainio',
    # Keywords that define your package best
    keywords=['decentralized', 'data', 'chainlink',
              'cryptocurrency', 'blockchain'],
    install_requires=[            # I get to this in a second
        'configparser',
        'pandas',
        'web3'
    ],
    test_requires=[
        'pytest',
        'configparser'
        'pandas',
        'web3'
    ],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    packages=find_packages(
        exclude=['test_chainlink_feeds', 'workflows', 'images']),
    # package_data={
    #     'chainlink_feeds': [],
    # }
)



# content of conftest.py
import pytest
import logging as log

from src.chainlink_feeds import ChainlinkFeeds
log.basicConfig(level=log.INFO)


@pytest.fixture
def get_test_abi():
    return '[{"inputs":[],"name":"decimals","outputs":[{"internalType":"uint8","name":"","type":"uint8"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"description","outputs":[{"internalType":"string","name":"","type":"string"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"uint80","name":"_roundId","type":"uint80"}],"name":"getRoundData","outputs":[{"internalType":"uint80","name":"roundId","type":"uint80"},{"internalType":"int256","name":"answer","type":"int256"},{"internalType":"uint256","name":"startedAt","type":"uint256"},{"internalType":"uint256","name":"updatedAt","type":"uint256"},{"internalType":"uint80","name":"answeredInRound","type":"uint80"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"latestRoundData","outputs":[{"internalType":"uint80","name":"roundId","type":"uint80"},{"internalType":"int256","name":"answer","type":"int256"},{"internalType":"uint256","name":"startedAt","type":"uint256"},{"internalType":"uint256","name":"updatedAt","type":"uint256"},{"internalType":"uint80","name":"answeredInRound","type":"uint80"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"version","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"}]'  # noqa


@pytest.fixture
def get_test_address():
    return '0x9326BFA02ADD2366b30bacB125260Af641031331'


# def test_rpc_url_not_present():
#     # Arrange
#     rpc_url = os.getenv('RPC_URL')
#     del os.environ['RPC_URL']

#     # Act / Assert
#     with pytest.raises(ValueError):
#         ChainlinkFeeds()

#     # reset
#     os.environ['RPC_URL'] = rpc_url


def test_get_latest_round_data():
    # Arrange
    network = 'kovan'
    pair = 'eth_usd'
    cf = ChainlinkFeeds(rpc_env_var='RPC_URL')

    # Act
    result = cf.get_latest_round_data(network=network, pair=pair)

    # Assert
    assert isinstance(result, dict)


def test_get_latest_round_data_with_abi_and_address(get_test_abi, get_test_address):
    # Arrange
    network = 'kovan'
    abi = get_test_abi
    address = get_test_address
    cf = ChainlinkFeeds(rpc_env_var='RPC_URL')

    # Act
    result = cf.get_latest_round_data(
        network=network, abi=abi, address=address)

    # Assert
    assert isinstance(result, dict)


def test_get_historical_price():
    # Arrange
    network = 'kovan'
    pair = 'eth_usd'
    round_id = 18446744073709556747
    cf = ChainlinkFeeds(rpc_env_var='RPC_URL')

    # Act
    result = cf.get_historical_price(round_id, network=network, pair=pair)

    # Assert
    assert isinstance(result, dict)


def test_load_config():
    # Arrange
    cf = ChainlinkFeeds(rpc_env_var='RPC_URL')

    # Act
    cf.load_config('./test/test_data/test_addresses_data.cfg',
                   abi_or_address='address')

    # Assert
    assert cf.get_manual_addresses(
    )['mainnet']['test_test'] == '0x0000000000000000000000000000000000000000'
